@@ -1,7 +1,16 @@
-/* $OpenLDAP: pkg/ldap/libraries/libldap/cyrus.c,v 1.45 2002/02/11 17:18:34 kurt Exp $ */
-/*
- * Copyright 1999-2002 The OpenLDAP Foundation, All Rights Reserved.
- * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
+/* $OpenLDAP: pkg/ldap/libraries/libldap/cyrus.c,v 1.88.2.18 2004/08/28 13:35:42 kurt Exp $ */
+/* This work is part of OpenLDAP Software <http://www.openldap.org/>.
+ *
+ * Copyright 1998-2004 The OpenLDAP Foundation.
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted only as authorized by the OpenLDAP
+ * Public License.
+ *
+ * A copy of this license is available in the file LICENSE in the
+ * top-level directory of the distribution or, alternatively, at
+ * <http://www.OpenLDAP.org/license.html>.
  */
 
 #include "portable.h"
@@ -14,19 +23,12 @@
 #include <ac/time.h>
 #include <ac/errno.h>
 #include <ac/ctype.h>
+#include <ac/unistd.h>
 
 #include "ldap-int.h"
 
 #ifdef HAVE_CYRUS_SASL
 
-#ifdef __APPLE__
-#include <pthread.h>
-static pthread_mutex_t _sasl_lock_data_ = PTHREAD_MUTEX_INITIALIZER;
-static pthread_mutex_t _sasl_lock_main_ = PTHREAD_MUTEX_INITIALIZER;
-static int _sasl_lock_count_ = 0;
-static unsigned int _sasl_thread_ = (unsigned int)-1;
-#endif
-
 #ifdef LDAP_R_COMPILE
 ldap_pvt_thread_mutex_t ldap_int_sasl_mutex;
 #endif
@@ -43,82 +45,62 @@ ldap_pvt_thread_mutex_t ldap_int_sasl_mutex;
 #define SASL_CONST
 #endif
 
-#ifdef __APPLE__
-void _lock_sasl(void)
-{
-	unsigned int t;
-
-	t = (unsigned int)pthread_self();
-
-	pthread_mutex_lock(&_sasl_lock_data_);
-	if ((_sasl_lock_count_ > 0) && (_sasl_thread_ == t))
-	{
-		/* Recursive lock: just increment the locked counter */
-		_sasl_lock_count_++;
-		pthread_mutex_unlock(&_sasl_lock_data_);
-		return;
-	}
-	pthread_mutex_unlock(&_sasl_lock_data_);
-
-	pthread_mutex_lock(&_sasl_lock_main_);
-
-	pthread_mutex_lock(&_sasl_lock_data_);
-	_sasl_lock_count_ = 1;
-	_sasl_thread_ = t;
-	pthread_mutex_unlock(&_sasl_lock_data_);
-}
-
-void _unlock_sasl(void)
-{
-	int unlock_me;
-
-	unlock_me = 0;
-
-	pthread_mutex_lock(&_sasl_lock_data_);
-	if (_sasl_lock_count_ > 0) _sasl_lock_count_--;
-	if (_sasl_lock_count_ == 0)
-	{
-		_sasl_thread_ = (unsigned int)-1;
-		unlock_me = 1;
-	}
-	pthread_mutex_unlock(&_sasl_lock_data_);
-
-	if (unlock_me == 1) pthread_mutex_unlock(&_sasl_lock_main_);
-}
-#endif
-
 /*
 * Various Cyrus SASL related stuff.
 */
 
+static const sasl_callback_t client_callbacks[] = {
+#ifdef SASL_CB_GETREALM
+	{ SASL_CB_GETREALM, NULL, NULL },
+#endif
+	{ SASL_CB_USER, NULL, NULL },
+	{ SASL_CB_AUTHNAME, NULL, NULL },
+	{ SASL_CB_PASS, NULL, NULL },
+	{ SASL_CB_ECHOPROMPT, NULL, NULL },
+	{ SASL_CB_NOECHOPROMPT, NULL, NULL },
+	{ SASL_CB_LIST_END, NULL, NULL }
+};
+
 int ldap_int_sasl_init( void )
 {
+	/* XXX not threadsafe */
 	static int sasl_initialized = 0;
 
-	static sasl_callback_t client_callbacks[] = {
-#ifdef SASL_CB_GETREALM
-		{ SASL_CB_GETREALM, NULL, NULL },
-#endif
-		{ SASL_CB_USER, NULL, NULL },
-		{ SASL_CB_AUTHNAME, NULL, NULL },
-		{ SASL_CB_PASS, NULL, NULL },
-		{ SASL_CB_ECHOPROMPT, NULL, NULL },
-		{ SASL_CB_NOECHOPROMPT, NULL, NULL },
-		{ SASL_CB_LIST_END, NULL, NULL }
-	};
+#ifdef HAVE_SASL_VERSION
+	/* stringify the version number, sasl.h doesn't do it for us */
+#define VSTR0(maj, min, pat)	#maj "." #min "." #pat
+#define VSTR(maj, min, pat)	VSTR0(maj, min, pat)
+#define SASL_VERSION_STRING	VSTR(SASL_VERSION_MAJOR, SASL_VERSION_MINOR, \
+				SASL_VERSION_STEP)
+	{ int rc;
+	sasl_version( NULL, &rc );
+	if ( ((rc >> 16) != ((SASL_VERSION_MAJOR << 8)|SASL_VERSION_MINOR)) ||
+		(rc & 0xffff) < SASL_VERSION_STEP) {
+		char version[sizeof("xxx.xxx.xxxxx")];
+		sprintf( version, "%u.%d.%d", (unsigned)rc >> 24, (rc >> 16) & 0xff,
+			rc & 0xffff );
 
-#ifdef __APPLE__
-	_lock_sasl();
+#ifdef NEW_LOGGING
+		LDAP_LOG( TRANSPORT, INFO,
+		"ldap_int_sasl_init: SASL library version mismatch:"
+		" expected " SASL_VERSION_STRING ","
+		" got %s\n", version, 0, 0 );
+#else
+		Debug( LDAP_DEBUG_ANY,
+		"ldap_int_sasl_init: SASL library version mismatch:"
+		" expected " SASL_VERSION_STRING ","
+		" got %s\n", version, 0, 0 );
+#endif
+		return -1;
+	}
+	}
 #endif
-
 	if ( sasl_initialized ) {
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return 0;
 	}
 
-#ifndef CSRIMALLOC
+/* SASL 2 takes care of its own memory completely internally */
+#if SASL_VERSION_MAJOR < 2 && !defined(CSRIMALLOC)
 	sasl_set_alloc(
 		ber_memalloc,
 		ber_memcalloc,
@@ -136,16 +118,14 @@ int ldap_int_sasl_init( void )
 	ldap_pvt_thread_mutex_init( &ldap_int_sasl_mutex );
 #endif
 
-	if ( sasl_client_init( client_callbacks ) == SASL_OK ) {
+	if ( sasl_client_init( NULL ) == SASL_OK ) {
 		sasl_initialized = 1;
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return 0;
 	}
 
-#ifdef __APPLE__
-	_unlock_sasl();
+#if SASL_VERSION_MAJOR < 2
+	/* A no-op to make sure we link with Cyrus 1.5 */
+	sasl_client_auth( NULL, NULL, NULL, 0, NULL, NULL );
 #endif
 	return -1;
 }
@@ -156,6 +136,7 @@ int ldap_int_sasl_init( void )
 
 struct sb_sasl_data {
 	sasl_conn_t		*sasl_context;
+	unsigned		*sasl_maxbuf;
 	Sockbuf_Buf		sec_buf_in;
 	Sockbuf_Buf		buf_in;
 	Sockbuf_Buf		buf_out;
@@ -166,38 +147,25 @@ sb_sasl_setup( Sockbuf_IO_Desc *sbiod, void *arg )
 {
 	struct sb_sasl_data	*p;
 
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
 	assert( sbiod != NULL );
 
 	p = LBER_MALLOC( sizeof( *p ) );
 	if ( p == NULL )
-	{
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return -1;
-	}
-
 	p->sasl_context = (sasl_conn_t *)arg;
 	ber_pvt_sb_buf_init( &p->sec_buf_in );
 	ber_pvt_sb_buf_init( &p->buf_in );
 	ber_pvt_sb_buf_init( &p->buf_out );
 	if ( ber_pvt_sb_grow_buffer( &p->sec_buf_in, SASL_MIN_BUFF_SIZE ) < 0 ) {
+		LBER_FREE( p );
 		errno = ENOMEM;
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return -1;
 	}
-
+	sasl_getprop( p->sasl_context, SASL_MAXOUTBUF,
+		(SASL_CONST void **) &p->sasl_maxbuf );
+	    
 	sbiod->sbiod_pvt = p;
 
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
 	return 0;
 }
 
@@ -206,10 +174,6 @@ sb_sasl_remove( Sockbuf_IO_Desc *sbiod )
 {
 	struct sb_sasl_data	*p;
 
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
 	assert( sbiod != NULL );
 	
 	p = (struct sb_sasl_data *)sbiod->sbiod_pvt;
@@ -226,11 +190,6 @@ sb_sasl_remove( Sockbuf_IO_Desc *sbiod )
 	ber_pvt_sb_buf_destroy( &p->buf_out );
 	LBER_FREE( p );
 	sbiod->sbiod_pvt = NULL;
-
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
-
 	return 0;
 }
 
@@ -239,10 +198,6 @@ sb_sasl_pkt_length( const unsigned char *buf, int debuglevel )
 {
 	ber_len_t		size;
 
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
 	assert( buf != NULL );
 
 	size = buf[0] << 24
@@ -250,9 +205,6 @@ sb_sasl_pkt_length( const unsigned char *buf, int debuglevel )
 		| buf[2] << 8
 		| buf[3];
    
-	/* we really should check against actual buffer size set
-	 * in the secopts.
-	 */
 	if ( size > SASL_MAX_BUFF_SIZE ) {
 		/* somebody is trying to mess me up. */
 		ber_log_printf( LDAP_DEBUG_ANY, debuglevel,
@@ -261,10 +213,6 @@ sb_sasl_pkt_length( const unsigned char *buf, int debuglevel )
 		size = 16; /* this should lead to an error. */
 	}
 
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
-
 	return size + 4; /* include the size !!! */
 }
 
@@ -274,27 +222,19 @@ sb_sasl_drop_packet ( Sockbuf_Buf *sec_buf_in, int debuglevel )
 {
 	ber_slen_t			len;
 
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
 	len = sec_buf_in->buf_ptr - sec_buf_in->buf_end;
 	if ( len > 0 )
 		AC_MEMCPY( sec_buf_in->buf_base, sec_buf_in->buf_base +
 			sec_buf_in->buf_end, len );
    
 	if ( len >= 4 ) {
-		sec_buf_in->buf_end = sb_sasl_pkt_length( sec_buf_in->buf_base,
-			debuglevel);
+		sec_buf_in->buf_end = sb_sasl_pkt_length(
+			(unsigned char *) sec_buf_in->buf_base, debuglevel);
 	}
 	else {
 		sec_buf_in->buf_end = 0;
 	}
 	sec_buf_in->buf_ptr = len;
-
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
 }
 
 static ber_slen_t
@@ -303,10 +243,6 @@ sb_sasl_read( Sockbuf_IO_Desc *sbiod, void *buf, ber_len_t len)
 	struct sb_sasl_data	*p;
 	ber_slen_t		ret, bufptr;
    
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
 	assert( sbiod != NULL );
 	assert( SOCKBUF_VALID( sbiod->sbiod_sb ) );
 
@@ -318,12 +254,7 @@ sb_sasl_read( Sockbuf_IO_Desc *sbiod, void *buf, ber_len_t len)
 	len -= ret;
 
 	if ( len == 0 )
-	{
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return bufptr;
-	}
 
 #if SASL_VERSION_MAJOR >= 2
 	ber_pvt_sb_buf_init( &p->buf_in );
@@ -333,25 +264,21 @@ sb_sasl_read( Sockbuf_IO_Desc *sbiod, void *buf, ber_len_t len)
 
 	/* Read the length of the packet */
 	while ( p->sec_buf_in.buf_ptr < 4 ) {
-		ret = LBER_SBIOD_READ_NEXT( sbiod, p->sec_buf_in.buf_base,
+		ret = LBER_SBIOD_READ_NEXT( sbiod, p->sec_buf_in.buf_base +
+			p->sec_buf_in.buf_ptr,
 			4 - p->sec_buf_in.buf_ptr );
 #ifdef EINTR
 		if ( ( ret < 0 ) && ( errno == EINTR ) )
 			continue;
 #endif
 		if ( ret <= 0 )
-		{
-#ifdef __APPLE__
-			_unlock_sasl();
-#endif
-			return ret;
-		}
+			return bufptr ? bufptr : ret;
 
 		p->sec_buf_in.buf_ptr += ret;
 	}
 
 	/* The new packet always starts at p->sec_buf_in.buf_base */
-	ret = sb_sasl_pkt_length( p->sec_buf_in.buf_base,
+	ret = sb_sasl_pkt_length( (unsigned char *) p->sec_buf_in.buf_base,
 		sbiod->sbiod_sb->sb_debug );
 
 	/* Grow the packet buffer if neccessary */
@@ -359,9 +286,6 @@ sb_sasl_read( Sockbuf_IO_Desc *sbiod, void *buf, ber_len_t len)
 		ber_pvt_sb_grow_buffer( &p->sec_buf_in, ret ) < 0 )
 	{
 		errno = ENOMEM;
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return -1;
 	}
 	p->sec_buf_in.buf_end = ret;
@@ -378,43 +302,36 @@ sb_sasl_read( Sockbuf_IO_Desc *sbiod, void *buf, ber_len_t len)
 			continue;
 #endif
 		if ( ret <= 0 )
-		{
-#ifdef __APPLE__
-			_unlock_sasl();
-#endif
-			return ret;
-		}
+			return bufptr ? bufptr : ret;
+
 		p->sec_buf_in.buf_ptr += ret;
    	}
 
 	/* Decode the packet */
-	ret = sasl_decode( p->sasl_context, p->sec_buf_in.buf_base,
-		p->sec_buf_in.buf_end,
-		(SASL_CONST char **)&p->buf_in.buf_base,
-		(unsigned *)&p->buf_in.buf_end );
+	{
+		unsigned tmpsize = p->buf_in.buf_end;
+		ret = sasl_decode( p->sasl_context, p->sec_buf_in.buf_base,
+			p->sec_buf_in.buf_end,
+			(SASL_CONST char **)&p->buf_in.buf_base,
+			(unsigned *)&tmpsize );
+		p->buf_in.buf_end = tmpsize;
+	}
+
+	/* Drop the packet from the input buffer */
+	sb_sasl_drop_packet( &p->sec_buf_in, sbiod->sbiod_sb->sb_debug );
+
 	if ( ret != SASL_OK ) {
 		ber_log_printf( LDAP_DEBUG_ANY, sbiod->sbiod_sb->sb_debug,
 			"sb_sasl_read: failed to decode packet: %s\n",
 			sasl_errstring( ret, NULL, NULL ) );
-		sb_sasl_drop_packet( &p->sec_buf_in,
-			sbiod->sbiod_sb->sb_debug );
 		errno = EIO;
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return -1;
 	}
 	
-	/* Drop the packet from the input buffer */
-	sb_sasl_drop_packet( &p->sec_buf_in, sbiod->sbiod_sb->sb_debug );
-
 	p->buf_in.buf_size = p->buf_in.buf_end;
 
 	bufptr += ber_pvt_sb_copy_out( &p->buf_in, (char*) buf + bufptr, len );
 
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
 	return bufptr;
 }
 
@@ -424,10 +341,6 @@ sb_sasl_write( Sockbuf_IO_Desc *sbiod, void *buf, ber_len_t len)
 	struct sb_sasl_data	*p;
 	int			ret;
 
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
 	assert( sbiod != NULL );
 	assert( SOCKBUF_VALID( sbiod->sbiod_sb ) );
 
@@ -436,12 +349,12 @@ sb_sasl_write( Sockbuf_IO_Desc *sbiod, void *buf, ber_len_t len)
 	/* Are there anything left in the buffer? */
 	if ( p->buf_out.buf_ptr != p->buf_out.buf_end ) {
 		ret = ber_pvt_sb_do_write( sbiod, &p->buf_out );
-		if ( ret <= 0 )
-		{
-#ifdef __APPLE__
-			_unlock_sasl();
-#endif
-			return ret;
+		if ( ret < 0 ) return ret;
+
+		/* Still have something left?? */
+		if ( p->buf_out.buf_ptr != p->buf_out.buf_end ) {
+			errno = EAGAIN;
+			return -1;
 		}
 	}
 
@@ -451,32 +364,32 @@ sb_sasl_write( Sockbuf_IO_Desc *sbiod, void *buf, ber_len_t len)
 #else
 	ber_pvt_sb_buf_destroy( &p->buf_out );
 #endif
-	ret = sasl_encode( p->sasl_context, buf, len,
-		(SASL_CONST char **)&p->buf_out.buf_base,
-		(unsigned *)&p->buf_out.buf_size );
+	if ( len > *p->sasl_maxbuf - 100 ) {
+		len = *p->sasl_maxbuf - 100;	/* For safety margin */
+	}
+
+	{
+		unsigned tmpsize = p->buf_out.buf_size;
+		ret = sasl_encode( p->sasl_context, buf, len,
+			(SASL_CONST char **)&p->buf_out.buf_base,
+			&tmpsize );
+		p->buf_out.buf_size = tmpsize;
+	}
+
 	if ( ret != SASL_OK ) {
 		ber_log_printf( LDAP_DEBUG_ANY, sbiod->sbiod_sb->sb_debug,
 			"sb_sasl_write: failed to encode packet: %s\n",
 			sasl_errstring( ret, NULL, NULL ) );
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
+		errno = EIO;
 		return -1;
 	}
 	p->buf_out.buf_end = p->buf_out.buf_size;
 
 	ret = ber_pvt_sb_do_write( sbiod, &p->buf_out );
-	if ( ret <= 0 )
-	{
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
-		return ret;
-	}
 
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
+	/* return number of bytes encoded, not written, to ensure
+	 * no byte is encoded twice (even if only sent once).
+	 */
 	return len;
 }
 
@@ -484,30 +397,14 @@ static int
 sb_sasl_ctrl( Sockbuf_IO_Desc *sbiod, int opt, void *arg )
 {
 	struct sb_sasl_data	*p;
-	int rc;
 
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
 	p = (struct sb_sasl_data *)sbiod->sbiod_pvt;
 
 	if ( opt == LBER_SB_OPT_DATA_READY ) {
-		if ( p->buf_in.buf_ptr != p->buf_in.buf_end )
-		{
-#ifdef __APPLE__
-			_unlock_sasl();
-#endif
-			return 1;
-		}
+		if ( p->buf_in.buf_ptr != p->buf_in.buf_end ) return 1;
 	}
-
-	rc = LBER_SBIOD_CTRL_NEXT( sbiod, opt, arg );
-
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
-
-	return rc;
+	
+	return LBER_SBIOD_CTRL_NEXT( sbiod, opt, arg );
 }
 
 Sockbuf_IO ldap_pvt_sockbuf_io_sasl = {
@@ -521,12 +418,12 @@ Sockbuf_IO ldap_pvt_sockbuf_io_sasl = {
 
 int ldap_pvt_sasl_install( Sockbuf *sb, void *ctx_arg )
 {
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
+#ifdef NEW_LOGGING
+	LDAP_LOG ( TRANSPORT, ENTRY, "ldap_pvt_sasl_install\n", 0, 0, 0 );
+#else
 	Debug( LDAP_DEBUG_TRACE, "ldap_pvt_sasl_install\n",
 		0, 0, 0 );
+#endif
 
 	/* don't install the stuff unless security has been negotiated */
 
@@ -541,22 +438,24 @@ int ldap_pvt_sasl_install( Sockbuf *sb, void *ctx_arg )
 			LBER_SBIOD_LEVEL_APPLICATION, ctx_arg );
 	}
 
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
-
 	return LDAP_SUCCESS;
 }
 
+void ldap_pvt_sasl_remove( Sockbuf *sb )
+{
+	ber_sockbuf_remove_io( sb, &ldap_pvt_sockbuf_io_sasl,
+		LBER_SBIOD_LEVEL_APPLICATION );
+#ifdef LDAP_DEBUG
+	ber_sockbuf_remove_io( sb, &ber_sockbuf_io_debug,
+		LBER_SBIOD_LEVEL_APPLICATION );
+#endif
+}
+
 static int
 sasl_err2ldap( int saslerr )
 {
 	int rc;
 
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
 	switch (saslerr) {
 		case SASL_CONTINUE:
 			rc = LDAP_MORE_RESULTS_TO_RETURN;
@@ -592,11 +491,6 @@ sasl_err2ldap( int saslerr )
 	}
 
 	assert( rc == LDAP_SUCCESS || LDAP_API_ERROR( rc ) );
-
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
-
 	return rc;
 }
 
@@ -604,105 +498,64 @@ int
 ldap_int_sasl_open(
 	LDAP *ld, 
 	LDAPConn *lc,
-	const char * host,
-	ber_len_t ssf )
+	const char * host )
 {
 	int rc;
 	sasl_conn_t *ctx;
-	sasl_callback_t *session_callbacks;
 
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
-	session_callbacks = LDAP_CALLOC( 2, sizeof( sasl_callback_t ) );
-
-	if( session_callbacks == NULL ) 
-	{
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
-		return LDAP_NO_MEMORY;
-	}
-
-	session_callbacks[0].id = SASL_CB_USER;
-	session_callbacks[0].proc = NULL;
-	session_callbacks[0].context = ld;
-
-	session_callbacks[1].id = SASL_CB_LIST_END;
-	session_callbacks[1].proc = NULL;
-	session_callbacks[1].context = NULL;
-
-	assert( lc->lconn_sasl_ctx == NULL );
+	assert( lc->lconn_sasl_authctx == NULL );
 
 	if ( host == NULL ) {
 		ld->ld_errno = LDAP_LOCAL_ERROR;
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
+		return ld->ld_errno;
+	}
+
+	if ( ldap_int_sasl_init() ) {
+		ld->ld_errno = LDAP_LOCAL_ERROR;
 		return ld->ld_errno;
 	}
 
 #if SASL_VERSION_MAJOR >= 2
 	rc = sasl_client_new( "ldap", host, NULL, NULL,
-		session_callbacks, 0, &ctx );
+		client_callbacks, 0, &ctx );
 #else
-	rc = sasl_client_new( "ldap", host, session_callbacks,
+	rc = sasl_client_new( "ldap", host, client_callbacks,
 		SASL_SECURITY_LAYER, &ctx );
 #endif
-	LDAP_FREE( session_callbacks );
 
 	if ( rc != SASL_OK ) {
 		ld->ld_errno = sasl_err2ldap( rc );
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return ld->ld_errno;
 	}
 
+#ifdef NEW_LOGGING
+	LDAP_LOG ( TRANSPORT, DETAIL1, "ldap_int_sasl_open: host=%s\n", 
+		host, 0, 0 );
+#else
 	Debug( LDAP_DEBUG_TRACE, "ldap_int_sasl_open: host=%s\n",
 		host, 0, 0 );
-
-	lc->lconn_sasl_ctx = ctx;
-
-	if( ssf ) {
-#if SASL_VERSION_MAJOR >= 2
-		(void) sasl_setprop( ctx, SASL_SSF_EXTERNAL,
-			(void *) &ssf );
-#else
-		sasl_external_properties_t extprops;
-		memset(&extprops, 0L, sizeof(extprops));
-		extprops.ssf = ssf;
-
-		(void) sasl_setprop( ctx, SASL_SSF_EXTERNAL,
-			(void *) &extprops );
 #endif
-		Debug( LDAP_DEBUG_TRACE, "ldap_int_sasl_open: ssf=%ld\n",
-			(long) ssf, 0, 0 );
-	}
 
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
+	lc->lconn_sasl_authctx = ctx;
+
 	return LDAP_SUCCESS;
 }
 
 int ldap_int_sasl_close( LDAP *ld, LDAPConn *lc )
 {
-	sasl_conn_t *ctx = lc->lconn_sasl_ctx;
-
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
+	sasl_conn_t *ctx = lc->lconn_sasl_authctx;
 
 	if( ctx != NULL ) {
 		sasl_dispose( &ctx );
-		lc->lconn_sasl_ctx = NULL;
+		if ( lc->lconn_sasl_sockctx &&
+			lc->lconn_sasl_authctx != lc->lconn_sasl_sockctx ) {
+			ctx = lc->lconn_sasl_sockctx;
+			sasl_dispose( &ctx );
+		}
+		lc->lconn_sasl_sockctx = NULL;
+		lc->lconn_sasl_authctx = NULL;
 	}
 
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
 	return LDAP_SUCCESS;
 }
 
@@ -722,25 +575,24 @@ ldap_int_sasl_bind(
 	const char *pmech = NULL;
 	int			saslrc, rc;
 	sasl_ssf_t		*ssf = NULL;
-	sasl_conn_t	*ctx;
+	sasl_conn_t	*ctx, *oldctx = NULL;
 	sasl_interact_t *prompts = NULL;
 	unsigned credlen;
 	struct berval ccred;
 	ber_socket_t		sd;
+	void	*ssl;
 
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
+#ifdef NEW_LOGGING
+	LDAP_LOG ( TRANSPORT, ARGS, "ldap_int_sasl_bind: %s\n", 
+		mechs ? mechs : "<null>", 0, 0 );
+#else
 	Debug( LDAP_DEBUG_TRACE, "ldap_int_sasl_bind: %s\n",
 		mechs ? mechs : "<null>", 0, 0 );
+#endif
 
 	/* do a quick !LDAPv3 check... ldap_sasl_bind will do the rest. */
 	if (ld->ld_version < LDAP_VERSION3) {
 		ld->ld_errno = LDAP_NOT_SUPPORTED;
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return ld->ld_errno;
 	}
 
@@ -751,35 +603,61 @@ ldap_int_sasl_bind(
  		int rc;
 
 		rc = ldap_open_defconn( ld );
-		if( rc < 0 )
-		{
-#ifdef __APPLE__
-			_unlock_sasl();
-#endif
-			return ld->ld_errno;
-		}
+		if( rc < 0 ) return ld->ld_errno;
 
-		ber_sockbuf_ctrl( ld->ld_sb, LBER_SB_OPT_GET_FD, &sd );
+		ber_sockbuf_ctrl( ld->ld_defconn->lconn_sb, LBER_SB_OPT_GET_FD, &sd );
 
 		if( sd == AC_SOCKET_INVALID ) {
 			ld->ld_errno = LDAP_LOCAL_ERROR;
-#ifdef __APPLE__
-			_unlock_sasl();
-#endif
 			return ld->ld_errno;
 		}
 	}   
 
-	ctx = ld->ld_defconn->lconn_sasl_ctx;
+	oldctx = ld->ld_defconn->lconn_sasl_authctx;
 
-	if( ctx == NULL ) {
-		ld->ld_errno = LDAP_LOCAL_ERROR;
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
-		return ld->ld_errno;
+	/* If we already have an authentication context, clear it out */
+	if( oldctx ) {
+		if ( oldctx != ld->ld_defconn->lconn_sasl_sockctx ) {
+			sasl_dispose( &oldctx );
+		}
+		ld->ld_defconn->lconn_sasl_authctx = NULL;
 	}
 
+	{ char *saslhost = ldap_host_connected_to( ld->ld_defconn->lconn_sb, "localhost" );
+	rc = ldap_int_sasl_open( ld, ld->ld_defconn, saslhost );
+	LDAP_FREE( saslhost );
+	}
+
+	if ( rc != LDAP_SUCCESS ) return rc;
+
+	ctx = ld->ld_defconn->lconn_sasl_authctx;
+
+	/* Check for TLS */
+	ssl = ldap_pvt_tls_sb_ctx( ld->ld_defconn->lconn_sb );
+	if ( ssl ) {
+		struct berval authid = BER_BVNULL;
+		ber_len_t fac;
+
+		fac = ldap_pvt_tls_get_strength( ssl );
+		/* failure is OK, we just can't use SASL EXTERNAL */
+		(void) ldap_pvt_tls_get_my_dn( ssl, &authid, NULL, 0 );
+
+		(void) ldap_int_sasl_external( ld, ld->ld_defconn, authid.bv_val, fac );
+		LDAP_FREE( authid.bv_val );
+	}
+
+#if !defined(_WIN32)
+	/* Check for local */
+	if ( ldap_pvt_url_scheme2proto( ld->ld_defconn->lconn_server->lud_scheme ) == LDAP_PROTO_IPC ) {
+		char authid[sizeof("uidNumber=4294967295+gidNumber=4294967295,"
+			"cn=peercred,cn=external,cn=auth")];
+		sprintf( authid, "uidNumber=%d+gidNumber=%d,"
+			"cn=peercred,cn=external,cn=auth",
+			(int) geteuid(), (int) getegid() );
+		(void) ldap_int_sasl_external( ld, ld->ld_defconn, authid, LDAP_PVT_SASL_LOCAL_SSF );
+	}
+#endif
+
 	/* (re)set security properties */
 	sasl_setprop( ctx, SASL_SEC_PROPS,
 		&ld->ld_options.ldo_sasl_secprops );
@@ -808,32 +686,23 @@ ldap_int_sasl_bind(
 			}
 		}
 
-#if SASL_VERSION_MAJOR >= 2
-		/* XXX the application should free interact results. */
-		if ( prompts != NULL && prompts->result != NULL ) {
-			LDAP_FREE( (void *)prompts->result );
-			prompts->result = NULL;
-		}
-#endif
-
 		if( saslrc == SASL_INTERACT ) {
 			int res;
 			if( !interact ) break;
 			res = (interact)( ld, flags, defaults, prompts );
-			if( res != LDAP_SUCCESS ) {
-				break;
-			}
+
+			if( res != LDAP_SUCCESS ) break;
 		}
 	} while ( saslrc == SASL_INTERACT );
 
 	ccred.bv_len = credlen;
 
 	if ( (saslrc != SASL_OK) && (saslrc != SASL_CONTINUE) ) {
-		ld->ld_errno = sasl_err2ldap( saslrc );
-#ifdef __APPLE__
-		_unlock_sasl();
+		rc = ld->ld_errno = sasl_err2ldap( saslrc );
+#if SASL_VERSION_MAJOR >= 2
+		ld->ld_error = LDAP_STRDUP( sasl_errdetail( ctx ) );
 #endif
-		return ld->ld_errno;
+		goto done;
 	}
 
 	do {
@@ -854,30 +723,37 @@ ldap_int_sasl_bind(
 		if ( rc != LDAP_SUCCESS && rc != LDAP_SASL_BIND_IN_PROGRESS ) {
 			if( scred && scred->bv_len ) {
 				/* and server provided us with data? */
+#ifdef NEW_LOGGING
+				LDAP_LOG ( TRANSPORT, DETAIL1, 
+					"ldap_int_sasl_bind: rc=%d sasl=%d len=%ld\n", 
+					rc, saslrc, scred->bv_len );
+#else
 				Debug( LDAP_DEBUG_TRACE,
 					"ldap_int_sasl_bind: rc=%d sasl=%d len=%ld\n",
 					rc, saslrc, scred->bv_len );
+#endif
 				ber_bvfree( scred );
 			}
-#ifdef __APPLE__
-			_unlock_sasl();
-#endif
-			return ld->ld_errno;
+			rc = ld->ld_errno;
+			goto done;
 		}
 
 		if( rc == LDAP_SUCCESS && saslrc == SASL_OK ) {
 			/* we're done, no need to step */
 			if( scred && scred->bv_len ) {
 				/* but server provided us with data! */
+#ifdef NEW_LOGGING
+				LDAP_LOG ( TRANSPORT, DETAIL1, 
+					"ldap_int_sasl_bind: rc=%d sasl=%d len=%ld\n", 
+					rc, saslrc, scred->bv_len );
+#else
 				Debug( LDAP_DEBUG_TRACE,
 					"ldap_int_sasl_bind: rc=%d sasl=%d len=%ld\n",
 					rc, saslrc, scred->bv_len );
-				ber_bvfree( scred );
-				ld->ld_errno = LDAP_LOCAL_ERROR;
-#ifdef __APPLE__
-				_unlock_sasl();
 #endif
-				return ld->ld_errno;
+				ber_bvfree( scred );
+				rc = ld->ld_errno = LDAP_LOCAL_ERROR;
+				goto done;
 			}
 			break;
 		}
@@ -890,24 +766,19 @@ ldap_int_sasl_bind(
 				(SASL_CONST char **)&ccred.bv_val,
 				&credlen );
 
+#ifdef NEW_LOGGING
+				LDAP_LOG ( TRANSPORT, DETAIL1, 
+					"ldap_int_sasl_bind: sasl_client_step: %d\n", saslrc,0,0 );
+#else
 			Debug( LDAP_DEBUG_TRACE, "sasl_client_step: %d\n",
 				saslrc, 0, 0 );
-
-#if SASL_VERSION_MAJOR >= 2
-			/* XXX the application should free interact results. */
-			if ( prompts != NULL && prompts->result != NULL ) {
-				LDAP_FREE( (void *)prompts->result );
-				prompts->result = NULL;
-			}
 #endif
 
 			if( saslrc == SASL_INTERACT ) {
 				int res;
 				if( !interact ) break;
 				res = (interact)( ld, flags, defaults, prompts );
-				if( res != LDAP_SUCCESS ) {
-					break;
-				}
+				if( res != LDAP_SUCCESS ) break;
 			}
 		} while ( saslrc == SASL_INTERACT );
 
@@ -916,26 +787,22 @@ ldap_int_sasl_bind(
 
 		if ( (saslrc != SASL_OK) && (saslrc != SASL_CONTINUE) ) {
 			ld->ld_errno = sasl_err2ldap( saslrc );
-#ifdef __APPLE__
-			_unlock_sasl();
+#if SASL_VERSION_MAJOR >= 2
+			ld->ld_error = LDAP_STRDUP( sasl_errdetail( ctx ) );
 #endif
-			return ld->ld_errno;
+			rc = ld->ld_errno;
+			goto done;
 		}
 	} while ( rc == LDAP_SASL_BIND_IN_PROGRESS );
 
-	if ( rc != LDAP_SUCCESS ) {
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
-		return rc;
-	}
+	if ( rc != LDAP_SUCCESS ) goto done;
 
 	if ( saslrc != SASL_OK ) {
-		ld->ld_errno = sasl_err2ldap( saslrc );
-#ifdef __APPLE__
-		_unlock_sasl();
+#if SASL_VERSION_MAJOR >= 2
+		ld->ld_error = LDAP_STRDUP( sasl_errdetail( ctx ) );
 #endif
-		return ld->ld_errno;
+		rc = ld->ld_errno = sasl_err2ldap( saslrc );
+		goto done;
 	}
 
 	if( flags != LDAP_SASL_QUIET ) {
@@ -944,14 +811,12 @@ ldap_int_sasl_bind(
 			fprintf( stderr, "SASL username: %s\n", data );
 		}
 
-#if SASL_VERSION_MAJOR >= 2
-		saslrc = sasl_getprop( ctx, SASL_DEFUSERREALM, (SASL_CONST void **) &data );
-#else
+#if SASL_VERSION_MAJOR < 2
 		saslrc = sasl_getprop( ctx, SASL_REALM, (SASL_CONST void **) &data );
-#endif
 		if( saslrc == SASL_OK && data && *data ) {
 			fprintf( stderr, "SASL realm: %s\n", data );
 		}
+#endif
 	}
 
 	saslrc = sasl_getprop( ctx, SASL_SSF, (SASL_CONST void **) &ssf );
@@ -965,13 +830,18 @@ ldap_int_sasl_bind(
 			if( flags != LDAP_SASL_QUIET ) {
 				fprintf( stderr, "SASL installing layers\n" );
 			}
+			if ( ld->ld_defconn->lconn_sasl_sockctx ) {
+				oldctx = ld->ld_defconn->lconn_sasl_sockctx;
+				sasl_dispose( &oldctx );
+				ldap_pvt_sasl_remove( ld->ld_defconn->lconn_sb );
+			}
 			ldap_pvt_sasl_install( ld->ld_conns->lconn_sb, ctx );
+			ld->ld_defconn->lconn_sasl_sockctx = ctx;
 		}
 	}
+	ld->ld_defconn->lconn_sasl_authctx = ctx;
 
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
+done:
 	return rc;
 }
 
@@ -988,21 +858,16 @@ ldap_int_sasl_external(
 	sasl_external_properties_t extprops;
 #endif
 
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
-	ctx = conn->lconn_sasl_ctx;
+	ctx = conn->lconn_sasl_authctx;
 
 	if ( ctx == NULL ) {
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return LDAP_LOCAL_ERROR;
 	}
    
 #if SASL_VERSION_MAJOR >= 2
 	sc = sasl_setprop( ctx, SASL_SSF_EXTERNAL, &ssf );
+	if ( sc == SASL_OK )
+		sc = sasl_setprop( ctx, SASL_AUTH_EXTERNAL, authid );
 #else
 	memset( &extprops, '\0', sizeof(extprops) );
 	extprops.ssf = ssf;
@@ -1013,15 +878,9 @@ ldap_int_sasl_external(
 #endif
 
 	if ( sc != SASL_OK ) {
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return LDAP_LOCAL_ERROR;
 	}
 
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
 	return LDAP_SUCCESS;
 }
 
@@ -1041,14 +900,7 @@ int ldap_pvt_sasl_secprops(
 	unsigned maxbufsize = 0;
 	int got_maxbufsize = 0;
 
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
 	if( props == NULL || secprops == NULL ) {
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return LDAP_PARAM_ERROR;
 	}
 
@@ -1083,39 +935,30 @@ int ldap_pvt_sasl_secprops(
 		} else if( !strncasecmp(props[i],
 			"minssf=", sizeof("minssf")) )
 		{
-			if( isdigit( props[i][sizeof("minssf")] ) ) {
+			if( isdigit( (unsigned char) props[i][sizeof("minssf")] ) ) {
 				got_min_ssf++;
 				min_ssf = atoi( &props[i][sizeof("minssf")] );
 			} else {
-#ifdef __APPLE__
-				_unlock_sasl();
-#endif
 				return LDAP_NOT_SUPPORTED;
 			}
 
 		} else if( !strncasecmp(props[i],
 			"maxssf=", sizeof("maxssf")) )
 		{
-			if( isdigit( props[i][sizeof("maxssf")] ) ) {
+			if( isdigit( (unsigned char) props[i][sizeof("maxssf")] ) ) {
 				got_max_ssf++;
 				max_ssf = atoi( &props[i][sizeof("maxssf")] );
 			} else {
-#ifdef __APPLE__
-				_unlock_sasl();
-#endif
 				return LDAP_NOT_SUPPORTED;
 			}
 
 		} else if( !strncasecmp(props[i],
 			"maxbufsize=", sizeof("maxbufsize")) )
 		{
-			if( isdigit( props[i][sizeof("maxbufsize")] ) ) {
+			if( isdigit( (unsigned char) props[i][sizeof("maxbufsize")] ) ) {
 				got_maxbufsize++;
 				maxbufsize = atoi( &props[i][sizeof("maxbufsize")] );
 			} else {
-#ifdef __APPLE__
-				_unlock_sasl();
-#endif
 				return LDAP_NOT_SUPPORTED;
 			}
 
@@ -1123,16 +966,10 @@ int ldap_pvt_sasl_secprops(
 				|| (maxbufsize > SASL_MAX_BUFF_SIZE )))
 			{
 				/* bad maxbufsize */
-#ifdef __APPLE__
-				_unlock_sasl();
-#endif
 				return LDAP_PARAM_ERROR;
 			}
 
 		} else {
-#ifdef __APPLE__
-			_unlock_sasl();
-#endif
 			return LDAP_NOT_SUPPORTED;
 		}
 	}
@@ -1151,9 +988,6 @@ int ldap_pvt_sasl_secprops(
 	}
 
 	ldap_charray_free( props );
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
 	return LDAP_SUCCESS;
 }
 
@@ -1162,42 +996,20 @@ ldap_int_sasl_config( struct ldapoptions *lo, int option, const char *arg )
 {
 	int rc;
 
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
 	switch( option ) {
 	case LDAP_OPT_X_SASL_SECPROPS:
 		rc = ldap_pvt_sasl_secprops( arg, &lo->ldo_sasl_secprops );
-		if( rc == LDAP_SUCCESS )
-		{
-#ifdef __APPLE__
-			_unlock_sasl();
-#endif
-			return 0;
-		}
+		if( rc == LDAP_SUCCESS ) return 0;
 	}
 
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
 	return -1;
 }
 
 int
 ldap_int_sasl_get_option( LDAP *ld, int option, void *arg )
 {
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
 	if ( ld == NULL )
-	{
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return -1;
-	}
 
 	switch ( option ) {
 		case LDAP_OPT_X_SASL_MECH: {
@@ -1223,18 +1035,12 @@ ldap_int_sasl_get_option( LDAP *ld, int option, void *arg )
 			sasl_conn_t *ctx;
 
 			if( ld->ld_defconn == NULL ) {
-#ifdef __APPLE__
-				_unlock_sasl();
-#endif
 				return -1;
 			}
 
-			ctx = ld->ld_defconn->lconn_sasl_ctx;
+			ctx = ld->ld_defconn->lconn_sasl_sockctx;
 
 			if ( ctx == NULL ) {
-#ifdef __APPLE__
-				_unlock_sasl();
-#endif
 				return -1;
 			}
 
@@ -1242,9 +1048,6 @@ ldap_int_sasl_get_option( LDAP *ld, int option, void *arg )
 				(SASL_CONST void **) &ssf );
 
 			if ( sc != SASL_OK ) {
-#ifdef __APPLE__
-				_unlock_sasl();
-#endif
 				return -1;
 			}
 
@@ -1253,9 +1056,6 @@ ldap_int_sasl_get_option( LDAP *ld, int option, void *arg )
 
 		case LDAP_OPT_X_SASL_SSF_EXTERNAL:
 			/* this option is write only */
-#ifdef __APPLE__
-			_unlock_sasl();
-#endif
 			return -1;
 
 		case LDAP_OPT_X_SASL_SSF_MIN:
@@ -1270,45 +1070,23 @@ ldap_int_sasl_get_option( LDAP *ld, int option, void *arg )
 
 		case LDAP_OPT_X_SASL_SECPROPS:
 			/* this option is write only */
-#ifdef __APPLE__
-			_unlock_sasl();
-#endif
 			return -1;
 
 		default:
-#ifdef __APPLE__
-			_unlock_sasl();
-#endif
 			return -1;
 	}
-
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
 	return 0;
 }
 
 int
 ldap_int_sasl_set_option( LDAP *ld, int option, void *arg )
 {
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
 	if ( ld == NULL )
-	{
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return -1;
-	}
 
 	switch ( option ) {
 	case LDAP_OPT_X_SASL_SSF:
 		/* This option is read-only */
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return -1;
 
 	case LDAP_OPT_X_SASL_SSF_EXTERNAL: {
@@ -1319,18 +1097,12 @@ ldap_int_sasl_set_option( LDAP *ld, int option, void *arg )
 		sasl_conn_t *ctx;
 
 		if( ld->ld_defconn == NULL ) {
-#ifdef __APPLE__
-			_unlock_sasl();
-#endif
 			return -1;
 		}
 
-		ctx = ld->ld_defconn->lconn_sasl_ctx;
+		ctx = ld->ld_defconn->lconn_sasl_authctx;
 
 		if ( ctx == NULL ) {
-#ifdef __APPLE__
-			_unlock_sasl();
-#endif
 			return -1;
 		}
 
@@ -1346,9 +1118,6 @@ ldap_int_sasl_set_option( LDAP *ld, int option, void *arg )
 #endif
 
 		if ( sc != SASL_OK ) {
-#ifdef __APPLE__
-			_unlock_sasl();
-#endif
 			return -1;
 		}
 		} break;
@@ -1368,98 +1137,70 @@ ldap_int_sasl_set_option( LDAP *ld, int option, void *arg )
 		sc = ldap_pvt_sasl_secprops( (char *) arg,
 			&ld->ld_options.ldo_sasl_secprops );
 
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return sc == LDAP_SUCCESS ? 0 : -1;
 		}
 
 	default:
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return -1;
 	}
-
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
 	return 0;
 }
 
 #ifdef LDAP_R_COMPILE
+#define LDAP_DEBUG_R_SASL
 void *ldap_pvt_sasl_mutex_new(void)
 {
 	ldap_pvt_thread_mutex_t *mutex;
 
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
 	mutex = (ldap_pvt_thread_mutex_t *) LDAP_MALLOC(
 		sizeof(ldap_pvt_thread_mutex_t) );
 
 	if ( ldap_pvt_thread_mutex_init( mutex ) == 0 ) {
-#ifdef __APPLE__
-		_unlock_sasl();
-#endif
 		return mutex;
 	}
-
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
+#ifndef LDAP_DEBUG_R_SASL
+	assert( 0 );
+#endif /* !LDAP_DEBUG_R_SASL */
 	return NULL;
 }
 
 int ldap_pvt_sasl_mutex_lock(void *mutex)
 {
-	int rc;
-
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
-	rc = ldap_pvt_thread_mutex_lock( (ldap_pvt_thread_mutex_t *)mutex )
+#ifdef LDAP_DEBUG_R_SASL
+	if ( mutex == NULL ) {
+		return SASL_OK;
+	}
+#else /* !LDAP_DEBUG_R_SASL */
+	assert( mutex );
+#endif /* !LDAP_DEBUG_R_SASL */
+	return ldap_pvt_thread_mutex_lock( (ldap_pvt_thread_mutex_t *)mutex )
 		? SASL_FAIL : SASL_OK;
-
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
-	return rc;
 }
 
 int ldap_pvt_sasl_mutex_unlock(void *mutex)
 {
-	int rc;
-
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
-	rc = ldap_pvt_thread_mutex_unlock( (ldap_pvt_thread_mutex_t *)mutex )
+#ifdef LDAP_DEBUG_R_SASL
+	if ( mutex == NULL ) {
+		return SASL_OK;
+	}
+#else /* !LDAP_DEBUG_R_SASL */
+	assert( mutex );
+#endif /* !LDAP_DEBUG_R_SASL */
+	return ldap_pvt_thread_mutex_unlock( (ldap_pvt_thread_mutex_t *)mutex )
 		? SASL_FAIL : SASL_OK;
-
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
-	return rc;
 }
 
 void ldap_pvt_sasl_mutex_dispose(void *mutex)
 {
-#ifdef __APPLE__
-	_lock_sasl();
-#endif
-
-#ifdef NOTDEF
+#ifdef LDAP_DEBUG_R_SASL
+	if ( mutex == NULL ) {
+		return;
+	}
+#else /* !LDAP_DEBUG_R_SASL */
+	assert( mutex );
+#endif /* !LDAP_DEBUG_R_SASL */
 	(void) ldap_pvt_thread_mutex_destroy( (ldap_pvt_thread_mutex_t *)mutex );
 	LDAP_FREE( mutex );
-#endif
-
-#ifdef __APPLE__
-	_unlock_sasl();
-#endif
 }
 #endif
 
